unit CmdExecThread;

(**
 *
 * The CmdExecThread unit contains TCmdExecThread to run a command
 * in a separate thread and TCmdExecComponent which is a TComponent
 * wrapper for TCmdExecThread.
 *
 * Unfortunately, the Libc.System() function does not work when integrated
 * debugging is enabled.  By default it is on, which is the way it should
 * be.  However you must turn it off when using this component to run an
 * application. (Tools, Debugger Options, left bottom corner).
 *
 * by Martin Lovell, martin@hunter-lovell.org
 * Freeware
 *
 *)

interface

uses
  Classes, SysUtils;

type

  (**
   * TCmdExecThread calls Libc.System() in a separate thread and
   * notifies the VCL thread when the function completes through an
   * event handler.
   *)
  TCmdExecThread = class(TThread)
  private
    { Private declarations }
  protected
    FOnComplete: TNotifyEvent;
    FCommand: String;
    FReturnValue: Integer;
    //FDone is used instead of Terminated because Terminated is
    // false in the OnComplete event handler.
    FDone: Boolean;
    procedure Execute; override;
    procedure FireCompleteEvent;
  public
    constructor Create;
    property OnComplete: TNotifyEvent read FOnComplete write FOnComplete;
    property Command: String read FCommand write FCommand;
    property ReturnValue: Integer read FReturnValue;
    property Done: Boolean read FDone write FDone;
  end;


  (**
   *
   * TCmdExecComponent is a TComponent wrapper for TCmdExecThread so that
   * it can be used from the VCL palette during design time.
   *
   *)
  TCmdExecComponent = class(TComponent)
  private
    FCommand: String;
    FFileName: String;
    FOnComplete: TNotifyEvent;
    FCmdThread: TCmdExecThread;
    function GetReturnValue: Integer;
    function GetDone: Boolean;
    procedure InternalComplete(Sender: TObject);
  protected
  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
    procedure RunCmd;
  published
    property OnComplete: TNotifyEvent read FOnComplete write FOnComplete;
    property Command: String read FCommand write FCommand;
    property ReturnValue: Integer read GetReturnValue;
    property Done: Boolean read GetDone;
    property FileName: String read FFileName write FFileName;
   end;

procedure Register;

implementation

uses
  Libc;

{ TCmdExecThread }

constructor TCmdExecThread.Create;
begin
  Inherited Create(True);
  FDone := False;
end;

procedure TCmdExecThread.Execute;
begin
  if FCommand <> '' then FReturnValue := Libc.System(PChar(FCommand));
  FDone := True;
  Synchronize(FireCompleteEvent);
end;

procedure TCmdExecThread.FireCompleteEvent;
begin
  if Assigned(FOnComplete) then FOnComplete(Self);
end;

{ TCmdExecComponent }
constructor TCmdExecComponent.Create(Owner: TComponent);
begin
  inherited;
  FCmdThread := nil;
end;

destructor TCmdExecComponent.Destroy;
begin
//  FCmdThread.Free;
//  inherited;
end;

function TCmdExecComponent.GetDone: Boolean;
begin
  if FCmdThread <> nil then Result := FCmdThread.Done
  else Result := False;
end;

function TCmdExecComponent.GetReturnValue: Integer;
begin
  if FCmdThread <> nil then Result := FCmdThread.ReturnValue
  else Result := 0; // raise Exception.Create('No return value is available');
end;


procedure TCmdExecComponent.InternalComplete(Sender: TObject);
begin
  if Assigned(FOnComplete) then FOnComplete(Self);
end;

procedure TCmdExecComponent.RunCmd;
begin
  if FCmdThread <> nil then
  begin
    if not FCmdThread.Done
      then exit; //raise Exception.Create('A command is already running')
//    else FCmdThread.Free;
  end;
  FCmdThread := TCmdExecThread.Create;
  FCmdThread.FreeOnTerminate := true;
  FCmdThread.OnComplete := InternalComplete;
  FCmdThread.Command := FCommand;
  FCmdThread.Resume;
end;


procedure Register;
begin
  RegisterComponents('Misc', [TCmdExecComponent]);
end;

end.
