//<FLAG License>
// Summary:
//   Directory Treeview component
//
// Description:
//   A Treeview that contains the accessible directories. Directories
//   are loaded on expanding. TreeNode.Data <> nil, if the according
//   subdirs are read.
//   Not very luxurious, but sufficient for JediDBD.
//   Platform independent (=> several compromises...)
//     Windows: Removable and cdrom drives are treated specially (read
//              not before selected and expanding)
//     Linux:   Symbolic links are checked whether they point to a directory
//
// Internal:
//
// Todo:
//   #ToDo1 further tests, documentation
//   #ToDo2 Refresh method
//   #ToDo3 UNC support on Windows
//
// Bugs:
//
// History:
// <pre>
//   <b>Version 00.01.00 (Released )</b>
//     04.10.2001 BBU  started
//     14.10.2001 BBU  Initial release
//     26.01.2002 BBU  SetDirectory implemented, symbolic links support (Linux)
// </pre>
//
//## License:
//##   The contents of this file are subject to the Mozilla Public License
//##   Version 1.1 (the "License"); you may not use this file except in
//##   compliance with the License. You may obtain a copy of the License at
//##   http://www.mozilla.org/MPL/
//##
//##   Software distributed under the License is distributed on an "AS IS"
//##   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//##   License for the specific language governing rights and limitations
//##   under the License.
//##
//##   The Original Code is JEDI Database Desktop
//##
//##   Initial Developer of the Original Code is JEDI Database Desktop Project
//##   Portions created by JEDI Database Desktop Project are
//##   Copyright  1997-2001 by JEDI Database Desktop Project.
//##   All Rights Reserved.
//##
//##   For a list of the contributors see the accompanying contributors.txt
//##   file.
//##
//##   You may retrieve the latest version of this file at the JEDI Database
//##   Desktop home page, located at http://jedidbd.sourceforge.net
//##
unit QDirectoryTreeview;

interface

uses
  Classes, QControls, QComCtrls, QFileListbox;

type
  TDirectoryTreeView = class(TCustomTreeView)
  private
    FFileListbox: TFileListbox;
    procedure SetDirectory(const Value: string);
    procedure SetFileListbox(const Value: TFileListbox);
    function GetDirectory: string;
  protected
    procedure Change(Node: TTreeNode); override;
    procedure ReadWindowsRoot;
    procedure ReadLinuxRoot;
    procedure ReadNextLevel(const Node: TTreeNode);
    function GetPath(Node: TTreeNode): string;
    procedure TreeViewExpanding(Sender: TObject; Node: TTreeNode;
      var AllowExpansion: Boolean);
    property ReadOnly default true;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Directory: string read GetDirectory write SetDirectory;
  published
    property FileListbox: TFileListbox read FFileListbox write SetFileListbox;

    property Align;
    property Anchors;
    property AutoExpand default False;
    property BorderStyle default bsSunken3d;
    property Color;
    property Constraints;
    property DragMode;
    property Enabled;
    property Font;
    property Height default 228;
    property Hint;
    property ItemMargin;
    property Indent;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property RowSelect;
    property ShowButtons;
    property ShowHint;
    property Sorted default true;
    property SortType default stText;
    property TabOrder;
    property TabStop;
    property Visible;
    property Width default 177;

    property OnChange;
    property OnChanging;
    property OnClick;
    property OnCollapsed;
    property OnCollapsing;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    // property OnExpanding;  used in this unit...
    property OnExpanded;
    property OnItemClick;
    property OnItemDoubleClick;
    property OnItemEnter;
    property OnItemExitViewportEnter;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyString;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property OnItemMouseDown;
    property OnViewPortMouseDown;
  end;

implementation

uses
  SysUtils, QForms
{$IFDEF MSWINDOWS}
  , Windows
{$ENDIF}
{$IFDEF LINUX}
  , Libc
{$ENDIF}
  ;
  
{ TDirectoryTreeView }

procedure TDirectoryTreeView.Change(Node: TTreeNode);
begin
  if Assigned(FFileListBox) then begin
    FFileListbox.Directory := Directory;;
  end;
  if Assigned(Node) then inherited Change(Node);
end;

constructor TDirectoryTreeView.Create(AOwner: TComponent);
begin
  inherited;
  Width := 177;
  Height := 228;
  SortType := stText;
  Sorted := True;
  ReadOnly := True;
  OnExpanding := TreeViewExpanding;

  {$IFDEF MSWINDOWS}
  //windows: read drive letters
  ReadWindowsRoot;
  {$ENDIF}

  {$IFDEF LINUX}
  //Linux: read root
  ReadLinuxRoot;
  {$ENDIF}

  if not (csDesigning in ComponentState) then Directory := GetCurrentDir;
end;

destructor TDirectoryTreeView.Destroy;
begin
  inherited;
end;

function TDirectoryTreeView.GetDirectory: string;
begin
  if Assigned(Selected) then begin
    Result := GetPath(Selected);
  end else begin
    Result := '';
  end;
end;

// returns full path to the Node (with trailing PathDelim)
function TDirectoryTreeView.GetPath(Node: TTreeNode): string;
begin
  Result := '';
  while Node.Parent <> nil do begin
    Result :=  Node.Text + PathDelim + Result;
    Node := Node.Parent;
  end;
  Result := Node.Text + Result;  // root has already trailing PathDelim
end;

procedure TDirectoryTreeView.ReadLinuxRoot;
{$IFDEF LINUX}
var
  NewNode: TTreeNode;
{$ENDIF}
begin
  {$IFDEF LINUX}
  NewNode := Items.Add(nil, PathDelim); { Add '/' }
  ReadNextLevel(NewNode);
  NewNode.Expand(False);
  Selected := NewNode;
  {$ENDIF}
end;

// read all subdirectories from node and add them to the treeview
procedure TDirectoryTreeView.ReadNextLevel(const Node: TTreeNode);
var
  Dir: String;
  sr: TSearchRec;
  PrevCursor: TCursor;
begin
  if (Node.Data = nil) and not (csDesigning in ComponentState) then begin
    PrevCursor:= Screen.Cursor;
    Screen.Cursor := crHourglass;
    Items.BeginUpdate;
    try
      Dir := GetPath(Node);
      if DirectoryExists(Dir) then begin
        sr.ExcludeAttr := 0;
        if FindFirst(Dir+'*', faAnyFile, sr) = 0 then begin
          repeat
            if ( ((sr.Attr and faDirectory) = faDirectory)
              {$IFDEF LINUX}
              {$WARN SYMBOL_PLATFORM OFF}
              or (((sr.Mode and S_IFLNK) > 0) and DirectoryExists(Dir+sr.Name))
              {$WARN SYMBOL_PLATFORM ON}
              {$ENDIF}
               )
              and (sr.Name <>'.') and (sr.Name<>'..')
            then begin
              Items.AddChild(Node, sr.Name);
            end;
          until FindNext(sr) <> 0;
          SysUtils.FindClose(sr);
        end; // if
      end;
    finally
      Node.Data := Pointer(1);
      Items.EndUpdate;
      Screen.Cursor := PrevCursor;
    end;
  end;
end;

procedure TDirectoryTreeView.ReadWindowsRoot;
{$IFDEF MSWINDOWS}
var
  DriveNum: Integer;
  DriveChar: Char;
  DriveBits: set of 0..25;
  DriveType: Integer;
  NewNode: TTreeNode;
  SelectNode: TTreeNode;
{$ENDIF}
begin
  {$IFDEF MSWINDOWS}
  //windows: read drive letters
  Integer(DriveBits) := GetLogicalDrives;
  SelectNode := nil;
  for DriveNum := 0 to 25 do begin
    if not (DriveNum in DriveBits) then Continue;
    DriveChar := Char(DriveNum + Ord('a'));
    DriveChar := Upcase(DriveChar);
    DriveType := GetDriveType(PChar(DriveChar + ':\'));
    case DriveType of
      // #ToDo3 use bitmaps for drive types
      DRIVE_REMOVABLE: begin
                         NewNode := Items.Add(nil, DriveChar +':' + PathDelim);
                         NewNode.HasChildren := True; // is set correctly in OnExpanding
                       end;
      DRIVE_FIXED    : begin
                         NewNode := Items.Add(nil, DriveChar +':' + PathDelim);
                         ReadNextLevel(NewNode);
                         if not Assigned(SelectNode) then SelectNode := NewNode;
                       end;
      DRIVE_REMOTE   : begin
                         NewNode := Items.Add(nil, DriveChar +':' + PathDelim);
                         ReadNextLevel(NewNode);
                         if not Assigned(SelectNode) then SelectNode := NewNode;
                       end;
      DRIVE_CDROM    : begin
                         NewNode := Items.Add(nil, DriveChar +':' + PathDelim);
                         NewNode.HasChildren := True; // is set correctly in OnExpanding 
                       end;
      DRIVE_RAMDISK  : begin
                         NewNode := Items.Add(nil, DriveChar +':' + PathDelim);
                         ReadNextLevel(NewNode);
                         if not Assigned(SelectNode) then SelectNode := NewNode;
                       end;
    end;
  end; // for
  // select first fixed/remote/ramdisk drive
  if Assigned(SelectNode) then Selected := SelectNode;
  {$ENDIF}
end;

// Tries to set the current directory to Value. Value is processed
// from left to right. If no matching directory exists, the tree descending
// stops at this point.
// Value is a fully qualified path name (with or without trailing PathDelim).
// Pathnames are case sensitive on Linux, but not on Windows!
procedure TDirectoryTreeView.SetDirectory(const Value: string);
var
  HeadPos: Integer;
  TailPos: Integer;
  CurPart: String;
  Dir: String;
  CurRoot: TTreeNode;
  CurNode: TTreeNode;
begin
  Items.BeginUpdate;
  CurRoot := nil;
  try
    Dir := IncludeTrailingPathDelimiter(Value);

    // Root nodes include PathDelim in their text; the others don't

    // Extract the root description (to the first PathDelim)
    HeadPos := Pos(PathDelim, Dir);
    TailPos := 1;
    CurPart := Copy(Dir, TailPos, HeadPos - TailPos + 1);
    Inc(HeadPos);

    // search the corresponding TreeNode
    CurRoot := Items.GetFirstNode;
    while Assigned(CurRoot) do begin
      if SameFileName(CurRoot.Text, CurPart) then Break;
      CurRoot := CurRoot.GetNextSibling;
    end; //while
    if not Assigned(CurRoot) then Exit;  // no matching TreeNode found

    while HeadPos < Length(Dir) do begin
      TailPos := HeadPos;

      if CurRoot.HasChildren then CurRoot.Expand(False) else Exit;

      // extract next part of Dir (without PathDelim)
      while (HeadPos < Length(Dir)) and (Dir[HeadPos] <> PathDelim) do Inc(HeadPos);
      CurPart := Copy(Dir, TailPos, HeadPos - TailPos);
      Inc(HeadPos);

      // search TreeNode
      CurNode := CurRoot.GetFirstChild;
      while Assigned(CurNode) do begin
        if SameFileName(CurNode.Text, CurPart) then Break;
        CurNode := CurRoot.GetNextChild(CurNode);
      end; // while CurNode <> nil
      if not Assigned(CurNode) then Exit;  // not found

      CurRoot := CurNode;
    end; // while HeadPos
  finally
    if Assigned(CurRoot) then begin
      CurRoot.MakeVisible;
      Selected := CurRoot;
    end;
    Items.EndUpdate;
  end;
end;

procedure TDirectoryTreeView.SetFileListbox(const Value: TFileListbox);
begin
  if FFileListbox <> Value then begin
    FFileListbox := Value;
    Change(nil);
  end;
end;

procedure TDirectoryTreeView.TreeViewExpanding(Sender: TObject;
  Node: TTreeNode; var AllowExpansion: Boolean);
var
  i: integer;  
begin
  Screen.Cursor := crHourglass;
  try
    {$IFDEF MSWINDOWS}
    // special case: lazy loading of removable/CD drives
    if Node.Data = nil then begin
      ReadNextLevel(Node);
      Node.HasChildren := (Node.Count > 0);
    end;
    {$ENDIF}

    // read the directories two levels deeper to show the next level correctly
    // (expanding buttons +)
    for i := 0 to Node.Count - 1 do begin
      ReadNextLevel(Node.Item[i]);
    end;
  finally
    AllowExpansion := True;
    Screen.Cursor := crDefault;
  end;
end;

end.
