//<FLAG License>
// Summary:
//   FileListbox-component
//
// Description:
//   A Listbox that contains the filenames of a defined directory.
//   Not very luxurious, but sufficient for JediDBD.
//   Platform independent (=> several compromises...)
//     Windows: Files with the archive flag set are included
//     Linux:   Symbolic links are examined whether they stand for a file or not
//
// Internal:
//
// Todo:
//   #ToDo1 further tests
//   #ToDo1 implement Options
//
// Bugs:
//
// History:
// <pre>
//   <b>Version 00.01.00 (Released )</b>
//     04.10.2001 BBU  started
//     14.10.2001 BBU  Initial release
//     26.01.2002 BBU  Symbolic links support (Linux)
// </pre>
//
//## License:
//##   The contents of this file are subject to the Mozilla Public License
//##   Version 1.1 (the "License"); you may not use this file except in
//##   compliance with the License. You may obtain a copy of the License at
//##   http://www.mozilla.org/MPL/
//##
//##   Software distributed under the License is distributed on an "AS IS"
//##   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//##   License for the specific language governing rights and limitations
//##   under the License.
//##
//##   The Original Code is JEDI Database Desktop
//##
//##   Initial Developer of the Original Code is JEDI Database Desktop Project
//##   Portions created by JEDI Database Desktop Project are
//##   Copyright  1997-2001 by JEDI Database Desktop Project.
//##   All Rights Reserved.
//##
//##   For a list of the contributors see the accompanying contributors.txt
//##   file.
//##
//##   You may retrieve the latest version of this file at the JEDI Database
//##   Desktop home page, located at http://jedidbd.sourceforge.net
//##
unit QFileListbox;

interface

uses
  Classes, QGraphics, QControls, QStdCtrls;

type
  // Options of the FileListbox
  TFileListBoxOption = (foShowReadonlyFiles {, foShowGlyphs, ...});
  TFileListBoxOptions = set of TFileListBoxOption;

  // The Filelistbox component.
  TFileListbox = class(TCustomListbox)
  private
    { Private declarations }
    FOptions: TFileListBoxOptions;
    procedure SetOptions(const Value: TFileListBoxOptions);
  protected
    { Protected declarations }
    FDirectory: String;
    FMask: string;
    FFiles: TStrings;
    procedure SetDirectory(const Value: string);
    procedure SetMask(const Value: string);
    procedure ReadFilenames; virtual;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  published
    { Published declarations }

    // On Windows: Directory contains also the drive letter
    // always with trailing PathDelim
    property Directory: string read FDirectory write SetDirectory;

    function FileName: string;

    function Files: TStrings;

    property Options: TFileListBoxOptions read FOptions write SetOptions;

    // Filter for display, e.g. '*.pas;*.dfm'
    // (a associated FilterEdit has to write this)
    property Mask: string read FMask write SetMask;

    property Height default 201;
    property Width default 177;

    property Style; { Must be published before Items }
    property Align;
    property Anchors;
    property BorderStyle;
    property Color default clBase;
    property Columns;
    property ColumnLayout;
    property Constraints;
    property DragMode;
    property Enabled;
    property ExtendedSelect;
    property Font;
    property ItemHeight;
    //property Items;     remains protected
    property MultiSelect;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property RowLayout;
    property Rows;
    property ShowHint default False;
    property Sorted default true;
    property TabOrder;
    property TabStop;
    property Visible;

    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnDrawItem;   // not tested; maybe it works!
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyString;
    property OnKeyUp;
    property OnMeasureItem; // not tested; maybe it works!
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;

  end;

implementation

uses
  SysUtils
  {$IFDEF LINUX}
  , Libc
  {$ENDIF}
  ;


{ TFileListbox }

constructor TFileListbox.Create(AOwner: TComponent);
begin
  inherited;
  Width := 177;
  Height := 201;
  Sorted := true;
  FOptions := [foShowReadOnlyFiles];
  FMask := '*';
  FFiles := TStringList.Create;
  Directory := GetCurrentDir;  // this triggers ReadFilenames
  ItemIndex := -1;
end;

destructor TFileListbox.Destroy;
begin
  FFiles.Free;
  inherited;
end;

// Returns full qualified filename (drive/dir/file/ext)
function TFileListbox.FileName: string;
begin
  if ItemIndex >= 0 then begin
    Result := FDirectory + Items[ItemIndex];
  end else begin
    Result := '';
  end;
end;

function TFileListbox.Files: TStrings;
var
  i: integer;
begin
  // #ToDo3 slow because FFiles is rebuilt on every access of Files
  FFiles.Clear;
  for i := 0 to Items.Count - 1 do begin
    if Selected[i] then FFiles.Add(FDirectory + Items[ItemIndex]);
  end;
  Result := FFiles;
end;

procedure TFileListbox.ReadFilenames;
var
  sr: TSearchRec;
  FileAttrs: Integer;
  tmpMask: String;      // unprocessed parts of the mask
  MaskEnd: integer;     // tmp
  MaskPart: String;     // currently processed part of the mask
begin
  Clear;
  if DirectoryExists(FDirectory) then begin
    {$IFDEF MSWINDOWS}
    {$WARN SYMBOL_PLATFORM OFF}
    FileAttrs := faArchive;
    if foShowReadOnlyFiles in Options then FileAttrs := FileAttrs or faReadOnly;
    {$WARN SYMBOL_PLATFORM ON}
    {$ENDIF}

    {$IFDEF LINUX}
    {$WARN SYMBOL_PLATFORM OFF}
    //#Todo2 what can we do on Linux?  rwx uga, sticky bit, ...???
    FileAttrs := 0;
    if foShowReadOnlyFiles in Options then FileAttrs := FileAttrs or faReadOnly;
    {$WARN SYMBOL_PLATFORM ON}
    {$ENDIF}

    //loop through all parts of the mask; they are separatey by ';'
    tmpMask := Mask;
    repeat
      MaskEnd := AnsiPos(';', tmpMask);
      if MaskEnd = 0 then MaskEnd := Length(tmpMask)+1;
      MaskPart := Trim(Copy(tmpMask, 1, MaskEnd-1));
      Delete(tmpMask, 1, MaskEnd);

      if FindFirst(FDirectory+MaskPart, FileAttrs, sr) = 0 then begin
        repeat
          if ((sr.Attr and FileAttrs) = sr.Attr)
          {$IFDEF LINUX}
          {$WARN SYMBOL_PLATFORM OFF}
              and (((sr.Mode and S_IFLNK) > 0)
                and not DirectoryExists(FDirectory+sr.Name))
          {$WARN SYMBOL_PLATFORM ON}
          {$ENDIF}
          then begin
            Items.Add(sr.Name);
          end;
        until FindNext(sr) <> 0;
        FindClose(sr);
      end; // if

    until Length(tmpMask) = 0;
  end; // if direxists
end;

procedure TFileListbox.SetDirectory(const Value: string);
begin
  FDirectory := Value;
  if length(FDirectory)=0 then FDirectory := PathDelim
  else if FDirectory[Length(FDirectory)] <> PathDelim then FDirectory := FDirectory + PathDelim;
  ReadFilenames;
end;


procedure TFileListbox.SetMask(const Value: string);
begin
  if FMask <> Value then begin
    if Length(Trim(Value))=0 then FMask :='*' else FMask := Value;
    ReadFilenames;
  end;
end;

procedure TFileListbox.SetOptions(const Value: TFileListBoxOptions);
begin
  FOptions := Value;
  ReadFilenames;
end;

end.
